# -*- coding: utf-8 -*-

import argparse
import logging
import sys

from s4 import VERSION
from s4 import utils
from s4.commands.add_command import AddCommand
from s4.commands.daemon_command import DaemonCommand
from s4.commands.edit_command import EditCommand
from s4.commands.ls_command import LsCommand
from s4.commands.rm_command import RmCommand
from s4.commands.sync_command import SyncCommand
from s4.commands.targets_command import TargetsCommand


def main(arguments):
    parser = argparse.ArgumentParser(
        prog="s4",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            "Fast and cheap synchronisation of files with Amazon S3\n"
            "\n"
            "Version: {}\n"
            "\n"
            'To start off, add a Target with the "add" command\n'
        ).format(VERSION),
    )
    parser.add_argument(
        "--log-level", default="INFO", choices=["DEBUG", "INFO", "WARNING", "ERROR"]
    )
    parser.add_argument(
        "--no-colors", action="store_true", help="Display without colors"
    )
    parser.add_argument(
        "--timestamps",
        action="store_true",
        help="Display timestamps for each log message",
    )
    subparsers = parser.add_subparsers(dest="command")

    daemon_parser = subparsers.add_parser(
        "daemon", help="Run S4 sync continiously", aliases=["d"]
    )
    daemon_parser.add_argument("targets", nargs="*")
    daemon_parser.add_argument("--read-delay", default=1000, type=int)
    daemon_parser.add_argument(
        "--conflicts", default="ignore", choices=["1", "2", "ignore"]
    )

    add_parser = subparsers.add_parser(
        "add", help="Add a new Target to synchronise", aliases=["a"]
    )
    add_parser.add_argument(
        "--copy-target-credentials",
        "-C",
        help="Copy credentials from an existing target instead of typing them in again",
    )

    sync_parser = subparsers.add_parser(
        "sync", help="Synchronise Targets with S3", aliases=["s"]
    )
    sync_parser.add_argument("targets", nargs="*")
    sync_parser.add_argument("--conflicts", default=None, choices=["1", "2", "ignore"])
    sync_parser.add_argument("--dry-run", action="store_true")

    edit_parser = subparsers.add_parser(
        "edit", help="Edit Target details", aliases=["e"]
    )
    edit_parser.add_argument("target")

    subparsers.add_parser("targets", help="Print available Targets", aliases=["t"])

    subparsers.add_parser("version", help="Print S4 Version", aliases=["v"])

    ls_parser = subparsers.add_parser("ls", help="Display list of files for a Target")
    ls_parser.add_argument("target")
    ls_parser.add_argument(
        "--sort-by", "-s", choices=["key", "local", "s3"], default="key"
    )
    ls_parser.add_argument("--descending", "-d", action="store_true")
    ls_parser.add_argument(
        "--all", "-A", dest="show_all", action="store_true", help="show deleted files"
    )

    remove_parser = subparsers.add_parser("rm", help="Remove a Target")
    remove_parser.add_argument("target")

    args = parser.parse_args(arguments)

    if args.log_level == "DEBUG":
        log_format = "%(levelname)s:%(module)s:%(lineno)s %(message)s"
    else:
        log_format = "%(message)s"

    if args.timestamps:
        log_format = "%(asctime)s: " + log_format

    logging.basicConfig(format=log_format, level=args.log_level)

    # shut boto up
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    logging.getLogger("botocore").setLevel(logging.CRITICAL)
    logging.getLogger("nose").setLevel(logging.CRITICAL)
    logging.getLogger("s3transfer").setLevel(logging.CRITICAL)
    logging.getLogger("filelock").setLevel(logging.CRITICAL)

    logger = logging.getLogger(__name__)
    logger.setLevel(args.log_level)

    config = utils.get_config()

    try:
        command = None
        if args.command in ("version", "v"):
            print(VERSION)
            return
        elif args.command in ("sync", "s"):
            command = SyncCommand(args, config, logger)
        elif args.command in ("targets", "t"):
            command = TargetsCommand(args, config, logger)
        elif args.command in ("add", "a"):
            command = AddCommand(args, config, logger)
        elif args.command in ("edit", "e"):
            command = EditCommand(args, config, logger)
        elif args.command == "ls":
            command = LsCommand(args, config, logger)
        elif args.command == "rm":
            command = RmCommand(args, config, logger)
        elif args.command in ("daemon", "d"):
            command = DaemonCommand(args, config, logger)

        if command:
            try:
                command.run()
            except Exception as e:
                logger.error("An unhandled error has occurred: %s", e)
                # Only display a scary stack trace to the user if in DEBUG mode
                if args.log_level == "DEBUG":
                    raise e
        else:
            parser.print_help()
    except KeyboardInterrupt:
        pass


def entry_point():
    main(sys.argv[1:])


if __name__ == "__main__":
    entry_point()



# -*- coding: utf-8 -*-

import datetime


class SyncState(object):
    UPDATED = "UPDATED"
    CREATED = "CREATED"
    DELETED = "DELETED"
    CONFLICT = "CONFLICT"
    NOCHANGES = "NOCHANGES"
    DOESNOTEXIST = "DOESNOTEXIST"

    def __init__(self, state, local_timestamp, remote_timestamp):
        self.state = state
        self.local_timestamp = local_timestamp
        self.remote_timestamp = remote_timestamp

    def get_local_datetime(self):
        if self.local_timestamp is not None:
            return datetime.datetime.utcfromtimestamp(self.local_timestamp)
        else:
            return None

    def get_remote_datetime(self):
        if self.remote_timestamp is not None:
            return datetime.datetime.utcfromtimestamp(self.remote_timestamp)
        else:
            return None

    def __eq__(self, other):
        if not isinstance(other, SyncState):
            return False
        return (
            self.state == other.state
            and self.local_timestamp == other.local_timestamp
            and self.remote_timestamp == other.remote_timestamp
        )

    def __repr__(self):
        return "SyncState<{}, local={}, remote={}>".format(
            self.state, self.get_local_datetime(), self.get_remote_datetime()
        )


class SyncObject(object):
    def __init__(self, fp, total_size, timestamp):
        self.fp = fp
        self.total_size = total_size
        self.timestamp = timestamp

    def __repr__(self):
        return "SyncObject<{}, {}, {}>".format(self.fp, self.total_size, self.timestamp)


def get_sync_state(index_local, real_local, remote):
    # convert to int because not all clients support float precision
    index_local = int(index_local) if index_local is not None else None
    real_local = int(real_local) if real_local is not None else None
    remote = int(remote) if remote is not None else None

    if index_local is None and real_local:
        return SyncState(SyncState.CREATED, real_local, remote)
    elif real_local is None and index_local:
        return SyncState(SyncState.DELETED, real_local, remote)
    elif real_local is None and index_local is None and remote:
        return SyncState(SyncState.DELETED, real_local, remote)
    elif real_local is None and index_local is None:
        # Does not exist in this case, so no action to perform
        return SyncState(SyncState.DOESNOTEXIST, None, None)

    elif index_local < real_local:
        return SyncState(SyncState.UPDATED, real_local, remote)
    elif index_local > real_local:
        return SyncState(SyncState.CONFLICT, index_local, remote)  # corruption?
    else:
        return SyncState(SyncState.NOCHANGES, real_local, remote)


class SyncClient(object):
    def get_client_name(self):
        """
        Return a human readable name for the client.
        """
        raise NotImplementedError()

    def get_uri(self, key=""):
        """
        Get the full Unique Resource Identifier given a key.
        """
        raise NotImplementedError()

    def lock(self, timeout=10):
        """
        Lock the client storage so that any other external access to the client is prevented.
        """
        raise NotImplementedError()

    def unlock(self):
        """
        Unlock the client storage for other external access. Will do nothing if nothing
        was locked in the first place.
        """
        raise NotImplementedError()

    def put(self, key, sync_object):
        """
        Put the given SyncObject on the clients storage with the given key.
        """
        raise NotImplementedError()

    def get(self, key):
        """
        Get a SyncObject reference for the given key. Returns None if it does
        not exist on the client's storage.
        """
        raise NotImplementedError()

    def delete(self, key):
        """
        Delete the given key from client storage.
        """
        raise NotImplementedError()

    def get_size(self, key):
        """
        Get the size of the file on the given client.
        """
        raise NotImplementedError()

    def get_local_keys(self):
        """
        Get *all* files that exists on the clients local storage. This means that
        keys outside the index could be returned if they are new. Also keys in the
        index could not be listed if the file has been deleted.
        """
        raise NotImplementedError()

    def get_real_local_timestamp(self, key):
        raise NotImplementedError()

    def get_index_keys(self):
        raise NotImplementedError()

    def get_index_local_timestamp(self, key):
        raise NotImplementedError()

    def set_index_local_timestamp(self, key, timestamp):
        raise NotImplementedError()

    def get_remote_timestamp(self, key):
        raise NotImplementedError()

    def set_remote_timestamp(self, key, timestamp):
        raise NotImplementedError()

    def get_all_remote_timestamps(self):
        raise NotImplementedError()

    def get_all_index_local_timestamps(self):
        raise NotImplementedError()

    def get_all_real_local_timestamps(self):
        raise NotImplementedError()

    def get_all_keys(self):
        local_keys = self.get_local_keys()
        index_keys = self.get_index_keys()
        return list(set(local_keys) | set(index_keys))

    def update_index(self):
        keys = self.get_all_keys()
        index = {}

        for key in keys:
            index[key] = {
                "remote_timestamp": self.get_remote_timestamp(key),
                "local_timestamp": self.get_real_local_timestamp(key),
            }
        self.index = index

    def update_index_entry(self, key):
        self.index[key] = {
            "remote_timestamp": self.get_remote_timestamp(key),
            "local_timestamp": self.get_real_local_timestamp(key),
        }

    def flush_index(self):
        raise NotImplementedError()

    def get_action(self, key):
        """
        returns the action to perform on this key based on its
        state before the last sync.
        """
        index_local_timestamp = self.get_index_local_timestamp(key)
        real_local_timestamp = self.get_real_local_timestamp(key)
        remote_timestamp = self.get_remote_timestamp(key)
        return get_sync_state(
            index_local_timestamp, real_local_timestamp, remote_timestamp
        )

    def get_all_actions(self):
        real_local_timestamps = self.get_all_real_local_timestamps()
        index_local_timestamps = self.get_all_index_local_timestamps()
        remote_timestamps = self.get_all_remote_timestamps()

        keys = set(real_local_timestamps) | set(index_local_timestamps)

        results = {}
        for key in keys:
            index_local_timestamp = index_local_timestamps.get(key)
            real_local_timestamp = real_local_timestamps.get(key)
            remote_timestamp = remote_timestamps.get(key)

            results[key] = get_sync_state(
                index_local_timestamp, real_local_timestamp, remote_timestamp
            )

        return results



# -*- encoding: utf-8 -*-

from s4.clients import SyncState


class Resolution(object):
    UPDATE = "UPDATE"
    CREATE = "CREATE"
    DELETE = "DELETE"

    def __init__(self, action, to_client, from_client, key, timestamp):
        self.action = action
        self.to_client = to_client
        self.from_client = from_client
        self.key = key
        self.timestamp = timestamp

    def __eq__(self, other):
        if not isinstance(other, Resolution):
            return False
        return (
            self.action == other.action
            and self.to_client == other.to_client
            and self.from_client == other.from_client
            and self.key == other.key
            and self.timestamp == other.timestamp
        )

    def __repr__(self):
        return "Resolution<action={}, to={}, from={}, key={}, timestamp={}>".format(
            self.action,
            self.to_client.get_uri() if self.to_client is not None else None,
            self.from_client.get_uri() if self.from_client is not None else None,
            self.key,
            self.timestamp,
        )

    @staticmethod
    def get_resolution(key, sync_state, to_client, from_client):
        if sync_state.state in (SyncState.UPDATED, SyncState.NOCHANGES):
            return Resolution(
                Resolution.UPDATE,
                to_client,
                from_client,
                key,
                sync_state.local_timestamp,
            )
        elif sync_state.state == SyncState.CREATED:
            return Resolution(
                Resolution.CREATE,
                to_client,
                from_client,
                key,
                sync_state.local_timestamp,
            )
        elif sync_state.state == SyncState.DELETED:
            return Resolution(
                Resolution.DELETE, to_client, None, key, sync_state.remote_timestamp
            )
        else:
            raise ValueError("Unknown action provided", sync_state)



#! -*- encoding: utf8 -*-
import os

from s4 import utils
from s4.commands import Command


class EditCommand(Command):
    def run(self):
        if "targets" not in self.config:
            self.logger.info("You have not added any targets yet")
            self.logger.info('Use the "add" command to do this')
            return

        if self.args.target not in self.config["targets"]:
            all_targets = sorted(list(self.config["targets"].keys()))
            self.logger.info('"%s" is an unknown target', self.args.target)
            self.logger.info("Choices are: %s", all_targets)
            return

        entry = self.config["targets"][self.args.target]

        local_folder = entry.get("local_folder", "")
        s3_uri = entry.get("s3_uri", "")
        aws_access_key_id = entry.get("aws_access_key_id")
        aws_secret_access_key = entry.get("aws_secret_access_key")
        region_name = entry.get("region_name")

        new_local_folder = utils.get_input("local folder [{}]: ".format(local_folder))
        new_s3_uri = utils.get_input("s3 uri [{}]: ".format(s3_uri))
        new_aws_access_key_id = utils.get_input(
            "AWS Access Key ID [{}]: ".format(aws_access_key_id)
        )

        secret_key_prompt = "AWS Secret Access Key [{}]: ".format(aws_secret_access_key)
        new_aws_secret_access_key = utils.get_input(secret_key_prompt, secret=True)

        new_region_name = utils.get_input("region name [{}]: ".format(region_name))

        if new_local_folder:
            entry["local_folder"] = os.path.expanduser(new_local_folder)
        if new_s3_uri:
            entry["s3_uri"] = new_s3_uri
        if new_aws_access_key_id:
            entry["aws_access_key_id"] = new_aws_access_key_id
        if new_aws_secret_access_key:
            entry["aws_secret_access_key"] = new_aws_secret_access_key
        if new_region_name:
            entry["region_name"] = new_region_name

        self.config["targets"][self.args.target] = entry

        utils.set_config(self.config)



#! -*- encoding: utf8 -*-

import os
import shutil
import subprocess
import tempfile


def show_diff(client_1, client_2, key):
    if shutil.which("diff") is None:
        print('Missing required "diff" executable.')
        print("Install this using your distribution's package manager")
        return

    if shutil.which("less") is None:
        print('Missing required "less" executable.')
        print("Install this using your distribution's package manager")
        return

    so1 = client_1.get(key)
    data1 = so1.fp.read()
    so1.fp.close()

    so2 = client_2.get(key)
    data2 = so2.fp.read()
    so2.fp.close()

    fd1, path1 = tempfile.mkstemp()
    fd2, path2 = tempfile.mkstemp()
    fd3, path3 = tempfile.mkstemp()

    with open(path1, "wb") as fp:
        fp.write(data1)
    with open(path2, "wb") as fp:
        fp.write(data2)

    # This is a lot faster than the difflib found in python
    with open(path3, "wb") as fp:
        subprocess.call(
            [
                "diff",
                "-u",
                "--label",
                client_1.get_uri(key),
                path1,
                "--label",
                client_2.get_uri(key),
                path2,
            ],
            stdout=fp,
        )

    subprocess.call(["less", path3])

    os.close(fd1)
    os.close(fd2)
    os.close(fd3)

    os.remove(path1)
    os.remove(path2)
    os.remove(path3)



#! -*- encoding: utf-8 -*-

from datetime import datetime

from tabulate import tabulate

from s4.commands import Command


class LsCommand(Command):
    def run(self):
        if "targets" not in self.config:
            self.logger.info("You have not added any targets yet")
            self.logger.info('Use the "add" command to do this')
            return
        if self.args.target not in self.config["targets"]:
            all_targets = sorted(list(self.config["targets"].keys()))
            self.logger.info('"%s" is an unknown target', self.args.target)
            self.logger.info("Choices are: %s", all_targets)
            return

        target = self.config["targets"][self.args.target]
        client_1, client_2 = self.get_clients(target)

        sort_by = self.args.sort_by.lower()
        descending = self.args.descending

        keys = set(client_1.get_index_keys()) | set(client_2.get_index_keys())

        total_size = 0

        data = []
        for key in sorted(keys):
            self.logger.debug("Processing %s", key)
            entry_1 = client_1.index.get(key, {})
            entry_2 = client_2.index.get(key, {})

            ts_1 = entry_1.get("local_timestamp")
            ts_2 = entry_2.get("local_timestamp")

            if self.args.show_all or ts_1 is not None:
                data.append(
                    (
                        key,
                        datetime.utcfromtimestamp(int(ts_1))
                        if ts_1 is not None
                        else "<deleted>",
                        datetime.utcfromtimestamp(int(ts_2))
                        if ts_2 is not None
                        else None,
                    )
                )
                size = client_1.get_size(key)
                self.logger.debug("%s size: %s", key, size)
                total_size += size

        headers = ["key", "local", "s3"]
        data = sorted(data, reverse=descending, key=lambda x: x[headers.index(sort_by)])

        print(tabulate(data, headers=headers))
        print("Total Size: {:.2f}Mb".format(total_size / (1024 * 1024)))



#! -*- encoding: utf8 -*
import os

from s4 import utils
from s4.commands import Command


class AddCommand(Command):
    def run(self):
        target = self.args.copy_target_credentials
        all_targets = list(self.config["targets"].keys())
        if target is not None and target not in all_targets:
            self.logger.info('"%s" is an unknown target', target)
            self.logger.info("Choices are: %s", all_targets)
            return

        entry = {}
        entry["local_folder"] = os.path.expanduser(utils.get_input("local folder: "))
        entry["s3_uri"] = utils.get_input("s3 uri: ")
        entry["region_name"] = utils.get_input("region name: ")

        if target is not None:
            entry["aws_access_key_id"] = self.config["targets"][target][
                "aws_access_key_id"
            ]
            entry["aws_secret_access_key"] = self.config["targets"][target][
                "aws_secret_access_key"
            ]
        else:
            entry["aws_access_key_id"] = utils.get_input("AWS Access Key ID: ")
            entry["aws_secret_access_key"] = utils.get_input(
                "AWS Secret Access Key: ", secret=True
            )

        default_name = os.path.basename(entry["s3_uri"])
        name = utils.get_input(
            "Provide a name for this entry [{}]: ".format(default_name)
        )

        if not name:
            name = default_name

        if "targets" not in self.config:
            self.config["targets"] = {}

        self.config["targets"][name] = entry

        utils.set_config(self.config)



#! -*- encoding: utf -*-

import sys

from clint.textui.colored import ColoredString

from s4 import sync, utils
from s4.commands import Command
from s4.diff import show_diff
from s4.progressbar import ProgressBar
from s4.resolution import Resolution


def handle_conflict(key, action_1, client_1, action_2, client_2):
    print(
        "\n"
        'Conflict for "{}". Which version would you like to keep?\n'
        "   (1) {}{} updated at {} ({})\n"
        "   (2) {}{} updated at {} ({})\n"
        "   (d) View difference (requires the diff command)\n"
        "   (X) Skip this file\n".format(
            key,
            client_1.get_uri(),
            key,
            action_1.get_remote_datetime(),
            action_1.state,
            client_2.get_uri(),
            key,
            action_2.get_remote_datetime(),
            action_2.state,
        ),
        file=sys.stderr,
    )
    while True:
        choice = utils.get_input("Choice (default=skip): ")
        print("", file=sys.stderr)

        if choice == "d":
            show_diff(client_1, client_2, key)
        else:
            break

    if choice == "1":
        return Resolution.get_resolution(key, action_1, client_2, client_1)
    elif choice == "2":
        return Resolution.get_resolution(key, action_2, client_1, client_2)


def display_progress_bar(sync_object):
    ProgressBar(
        total=sync_object.total_size,
        leave=False,
        ncols=80,
        unit="B",
        unit_scale=True,
        mininterval=0.2,
    )


def update_progress_bar(value):
    ProgressBar.update(value)


def hide_progress_bar(sync_object):
    ProgressBar.close()


class SyncCommand(Command):
    def run(self):
        all_targets = list(self.config["targets"].keys())
        if not self.args.targets:
            targets = all_targets
        else:
            targets = self.args.targets

        try:
            for name in sorted(targets):
                if name not in self.config["targets"]:
                    self.logger.info(
                        '"%s" is an unknown target. Choices are: %s', name, all_targets
                    )
                    continue

                entry = self.config["targets"][name]
                client_1, client_2 = self.get_clients(entry)

                try:
                    worker = sync.SyncWorker(
                        client_1,
                        client_2,
                        start_callback=display_progress_bar,
                        update_callback=update_progress_bar,
                        complete_callback=hide_progress_bar,
                        conflict_handler=handle_conflict,
                        action_callback=self.action_callback,
                    )

                    self.logger.info(
                        "Syncing %s [%s <=> %s]",
                        name,
                        client_1.get_uri(),
                        client_2.get_uri(),
                    )
                    worker.sync(
                        conflict_choice=self.args.conflicts, dry_run=self.args.dry_run
                    )
                except Exception as e:
                    if self.args.log_level == "DEBUG":
                        self.logger.exception(e)
                    else:
                        self.logger.error(
                            "There was an error syncing '%s': %s", name, e
                        )

        except KeyboardInterrupt:
            self.logger.warning("Quitting due to Keyboard Interrupt...")

    def action_callback(self, resolution):
        if resolution.action == Resolution.UPDATE:
            self.logger.info(
                self._colored("YELLOW", "Updating %s (%s => %s)"),
                resolution.key,
                resolution.from_client.get_uri(),
                resolution.to_client.get_uri(),
            )
        elif resolution.action == Resolution.CREATE:
            self.logger.info(
                self._colored("GREEN", "Creating %s (%s => %s)"),
                resolution.key,
                resolution.from_client.get_uri(),
                resolution.to_client.get_uri(),
            )
        elif resolution.action == Resolution.DELETE:
            self.logger.info(
                self._colored("RED", "Deleting %s on %s"),
                resolution.key,
                resolution.to_client.get_uri(),
            )

    def _colored(self, color, text):
        return text if self.args.no_colors else ColoredString(color, text)



# -*- coding: utf-8 -*-

import datetime
import getpass
import json
import os


CONFIG_FOLDER_PATH = os.path.expanduser("~/.config/s4")
CONFIG_FILE_PATH = os.path.join(CONFIG_FOLDER_PATH, "sync.conf")


def to_timestamp(dt):
    epoch = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
    return (dt - epoch) / datetime.timedelta(seconds=1)


def get_input(*args, secret=False, **kwargs):
    if secret:
        return getpass.getpass(*args, **kwargs)
    else:
        return input(*args, **kwargs)


def get_config():
    if not os.path.exists(CONFIG_FILE_PATH):
        return {"targets": {}}

    with open(CONFIG_FILE_PATH, "r") as fp:
        config = json.load(fp)
    return config


def set_config(config):
    if not os.path.exists(CONFIG_FOLDER_PATH):
        os.makedirs(CONFIG_FOLDER_PATH)

    with open(CONFIG_FILE_PATH, "w") as fp:
        json.dump(config, fp)



#! -*- encoding: utf8 -*-

from s4 import sync
from s4.clients.local import get_local_client
from s4.clients.s3 import get_s3_client


class Command(object):
    def __init__(self, args, config, logger):
        self.args = args
        self.config = config
        self.logger = logger

    def get_sync_worker(self, target):
        entry = self.config["targets"][target]
        client_1, client_2 = self.get_clients(entry)
        return sync.SyncWorker(client_1, client_2)

    def get_clients(self, entry):
        target_1 = entry["local_folder"]
        target_2 = entry["s3_uri"]
        aws_access_key_id = entry["aws_access_key_id"]
        aws_secret_access_key = entry["aws_secret_access_key"]
        region_name = entry["region_name"]

        # append trailing slashes to prevent incorrect prefix matching on s3
        if not target_1.endswith("/"):
            target_1 += "/"
        if not target_2.endswith("/"):
            target_2 += "/"

        client_1 = get_local_client(target_1)
        client_2 = get_s3_client(
            target_2, aws_access_key_id, aws_secret_access_key, region_name
        )
        return client_1, client_2



#! -*- encoding: utf-8 -*-

import tqdm


class ProgressBar(object):
    """
    Singleton wrapper around tqdm
    """

    pbar = None

    def __new__(cls, *args, **kwargs):
        if cls.pbar:
            cls.pbar.close()

        cls.pbar = tqdm.tqdm(*args, **kwargs)

    @classmethod
    def update(cls, value):
        cls.pbar.update(value)

    @classmethod
    def close(cls):
        cls.pbar.close()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['s4', 's4.clients', 's4.commands']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.7,<2.0',
 'clint>=0.5.1,<0.6.0',
 'filelock>=3.0,<4.0',
 'inotify-simple>=1.1,<2.0',
 'pathspec>=0.5.6,<0.6.0',
 'python-magic>=0.4.15,<0.5.0',
 'tabulate>=0.8.2,<0.9.0',
 'tqdm>=4.23,<5.0']

entry_points = \
{'console_scripts': ['s4 = s4.cli:entry_point']}

setup_kwargs = {
    'name': 's4',
    'version': '0.2.19',
    'description': 'Fast and cheap synchronisation of files using Amazon S3',
    'long_description': None,
    'author': 'Michael Aquilina',
    'author_email': 'michaelaquilina@gmail.com',
    'url': 'https://github.com/MichaelAquilina/S4',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)



#! -*- encoding: utf8 -*-

from s4.commands import Command


class TargetsCommand(Command):
    def run(self):
        for name in sorted(self.config["targets"]):
            entry = self.config["targets"][name]
            print(
                "{}: [{} <=> {}]".format(name, entry["local_folder"], entry["s3_uri"])
            )



# -*- coding: utf-8 -*-
import collections
import copy
import fnmatch
import gzip
import json
import logging
import os
import zlib

import boto3

from botocore.exceptions import ClientError

import magic

from s4 import utils
from s4.clients import SyncClient, SyncObject


logger = logging.getLogger(__name__)


S3Uri = collections.namedtuple("S3Uri", ["bucket", "key"])


def get_s3_client(target, aws_access_key_id, aws_secret_access_key, region_name):
    s3_uri = parse_s3_uri(target)
    s3_client = boto3.client(
        "s3",
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
    )
    return S3SyncClient(s3_client, s3_uri.bucket, s3_uri.key)


def parse_s3_uri(uri):
    if not uri.startswith("s3://"):
        return None

    tokens = uri.replace("s3://", "").split("/")
    if len(tokens) < 2:
        return None

    bucket = tokens[0]
    key = "/".join(tokens[1:])
    return S3Uri(bucket, key)


def is_ignored_key(key, ignore_files):
    # Check if any subdirectories match the ignore patterns
    key_parts = key.split("/")
    for part in key_parts:
        if any(fnmatch.fnmatch(part, pattern) for pattern in ignore_files):
            return True
    else:
        return False


class S3SyncClient(SyncClient):
    DEFAULT_IGNORE_FILES = [".index", ".s4lock"]

    def __init__(self, boto, bucket, prefix):
        self.boto = boto
        self.bucket = bucket
        self.prefix = prefix
        # These are lazy loaded as needed
        self._index = None
        self._ignore_files = None

    def lock(self):
        pass

    def unlock(self):
        pass

    def get_client_name(self):
        return "s3"

    def __repr__(self):
        return "S3SyncClient<{}, {}>".format(self.bucket, self.prefix)

    def get_uri(self, key=""):
        return "s3://{}/{}".format(self.bucket, os.path.join(self.prefix, key))

    def index_path(self):
        return os.path.join(self.prefix, ".index")

    @property
    def index(self):
        if self._index is None:
            self._index = self.load_index()
        return self._index

    @index.setter
    def index(self, value):
        self._index = value

    def put(self, key, sync_object, callback=None):
        self.boto.upload_fileobj(
            Bucket=self.bucket,
            Key=os.path.join(self.prefix, key),
            Fileobj=sync_object.fp,
            Callback=callback,
        )
        self.set_remote_timestamp(key, sync_object.timestamp)

    def get(self, key):
        try:
            resp = self.boto.get_object(
                Bucket=self.bucket, Key=os.path.join(self.prefix, key)
            )
            return SyncObject(
                resp["Body"],
                resp["ContentLength"],
                utils.to_timestamp(resp["LastModified"]),
            )
        except ClientError:
            return None

    def delete(self, key):
        resp = self.boto.delete_objects(
            Bucket=self.bucket,
            Delete={"Objects": [{"Key": os.path.join(self.prefix, key)}]},
        )
        return "Deleted" in resp

    def load_index(self):
        try:
            resp = self.boto.get_object(Bucket=self.bucket, Key=self.index_path())
            body = resp["Body"].read()
            content_type = magic.from_buffer(body, mime=True)

            if content_type == "text/plain":
                logger.debug("Detected plain text encoding for index")
                return json.loads(body.decode("utf-8"))

            # the magic/file command reports gzip differently depending on its version
            elif content_type in ("application/x-gzip", "application/gzip"):
                logger.debug("Detected gzip encoding for index")
                body = gzip.decompress(body)
                return json.loads(body.decode("utf-8"))

            # Older versions of Ubuntu cannot detect zlib files
            # assume octet-stream is zlib.
            # If it isnt, the decompress function will blow up anyway
            elif content_type in ("application/zlib", "application/octet-stream"):
                logger.debug("Detected zlib encoding for index")
                body = zlib.decompress(body)
                return json.loads(body.decode("utf-8"))

            elif content_type == "application/x-empty":
                return {}
            else:
                raise ValueError("Unknown content type for index", content_type)
        except (ClientError):
            return {}

    def reload_index(self):
        self.index = self.load_index()

    def flush_index(self, compressed=True):
        data = json.dumps(self.index).encode("utf-8")
        if compressed:
            logger.debug("Using gzip encoding for writing index")
            data = gzip.compress(data)
        else:
            logger.debug("Using plain text encoding for writing index")

        self.boto.put_object(Bucket=self.bucket, Key=self.index_path(), Body=data)

    def get_local_keys(self):
        results = []
        paginator = self.boto.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=self.bucket, Prefix=self.prefix)
        for page in page_iterator:
            if "Contents" not in page:
                return results

            for obj in page["Contents"]:
                key = os.path.relpath(obj["Key"], self.prefix)
                if not is_ignored_key(key, self.ignore_files):
                    results.append(key)
                else:
                    logger.debug("Ignoring %s", key)

        return results

    def get_real_local_timestamp(self, key):
        try:
            response = self.boto.head_object(
                Bucket=self.bucket, Key=os.path.join(self.prefix, key)
            )
            return utils.to_timestamp(response["LastModified"])
        except ClientError:
            return None

    def get_index_keys(self):
        return self.index.keys()

    def get_index_local_timestamp(self, key):
        return self.index.get(key, {}).get("local_timestamp")

    def set_index_local_timestamp(self, key, timestamp):
        if key not in self.index:
            self.index[key] = {}
        self.index[key]["local_timestamp"] = timestamp

    def get_remote_timestamp(self, key):
        return self.index.get(key, {}).get("remote_timestamp")

    def set_remote_timestamp(self, key, timestamp):
        if key not in self.index:
            self.index[key] = {}
        self.index[key]["remote_timestamp"] = timestamp

    def get_all_real_local_timestamps(self):
        result = {}
        paginator = self.boto.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=self.bucket, Prefix=self.prefix)
        for page in page_iterator:
            for obj in page.get("Contents", []):
                key = os.path.relpath(obj["Key"], self.prefix)
                if not is_ignored_key(key, self.ignore_files):
                    result[key] = utils.to_timestamp(obj["LastModified"])

        return result

    def get_all_remote_timestamps(self):
        return {key: value.get("remote_timestamp") for key, value in self.index.items()}

    def get_all_index_local_timestamps(self):
        return {key: value.get("local_timestamp") for key, value in self.index.items()}

    @property
    def ignore_files(self):
        if self._ignore_files is None:
            self.reload_ignore_files()
        return self._ignore_files

    def reload_ignore_files(self):
        self._ignore_files = copy.copy(self.DEFAULT_IGNORE_FILES)
        try:
            response = self.boto.get_object(
                Bucket=self.bucket, Key=os.path.join(self.prefix, ".syncignore")
            )
            data = response["Body"].read()
            data = data.decode("utf8")
            ignore_list = data.split("\n")
            self._ignore_files.extend(ignore_list)
        except ClientError:
            pass



#! -*- encoding: utf-8 -*-

from s4 import utils
from s4.commands import Command


class RmCommand(Command):
    def run(self):
        if "targets" not in self.config:
            self.logger.info("You have not added any targets yet")
            return
        if self.args.target not in self.config["targets"]:
            all_targets = sorted(list(self.config["targets"].keys()))
            self.logger.info('"%s" is an unknown target', self.args.target)
            self.logger.info("Choices are: %s", all_targets)
            return

        del self.config["targets"][self.args.target]
        utils.set_config(self.config)
